from interface import Interface


class Audio(Interface):
    def read(self):
        pass

    def getAudioFileName(self):
        pass

    def findIntervals(self, smoothingWindow, weight, plot):
        pass



import numpy as np
from scipy.io import wavfile


def read_audio_file(input_file):
    """
    This function returns a numpy array that stores the audio samples of a
    specified WAV of AIFF file
    """

    sampling_rate = 0
    signal = np.array([])
    if isinstance(input_file, str):
        sampling_rate, signal = wavfile.read(input_file)

    if signal.ndim == 2 and signal.shape[1] == 1:
        signal = signal.flatten()

    return sampling_rate, signal


def stereo_to_mono(signal):
    """
    This function converts the input signal
    (stored in a numpy array) to MONO (if it is STEREO)
    """

    if signal.ndim == 2:
        if signal.shape[1] == 1:
            signal = signal.flatten()
        else:
            if signal.shape[1] == 2:
                signal = (signal[:, 1] / 2) + (signal[:, 0] / 2)
    return signal



from config import DEFAULT_SAMPLING_RATE, AUDIO_DIR, VIDEO_DIR
from interface import implements
from AudioExtraction.Video import Video
import subprocess
import os
from Utilities import Utilities as util


class Mp4Video(implements(Video)):
    def __init__(self, videoFileName, samplingRate=DEFAULT_SAMPLING_RATE):
        if len(videoFileName.split('/')) > 1:
            raise Exception(
                "Invalid video file name. Valid example: 001-01.mp4")

        if not videoFileName.endswith('.mp4'):
            raise TypeError(
                "{} is not a mp4 video format".format(videoFileName))

        self.videoFileName = videoFileName
        self.samplingRate = samplingRate
        self.videoFilePath = util.getVideoPath(self.videoFileName)
        self.audioFilePath = util.getAudioFilePath(self.videoFileName)

        if not os.path.isfile(self.videoFilePath):
            raise FileNotFoundError(
                "Video file '{}' not does not exists".format(self.videoFilePath))

    def extract(self):
        command = "ffmpeg -loglevel quiet -i {} -ab 160k -ac 2 -ar {} -vn {}".format(
            self.videoFilePath, self.samplingRate, self.audioFilePath)

        subprocess.call(command, shell=True)



import config
import os


def getSpeakerBatch(fileName):
    speaker = fileName[0:config.SPEAKER_ID_LENGTH]
    batch = fileName[config.SPEAKER_ID_LENGTH +
                     1: config.SPEAKER_ID_LENGTH + config.BATCH_ID_LENGTH + 1]
    return speaker, batch


def getIntervalPath(fileName):
    return os.path.join(config.INTERVAL_DIR, fileName.split('.')[0] + '.csv')


def getVideoPath(fileName):
    return os.path.join(config.VIDEO_DIR, fileName.split('.')[0] + '.mp4')


def getAudioFilePath(fileName):
    return os.path.join(config.AUDIO_DIR, fileName.split('.')[0] + '.wav')


def getUncompressedVideoPath(fileName):
    return os.path.join(config.UNCOMPRESSED_VIDEO_DIR, fileName.split('.')[0] + '.mp4')


def getSeparatedAudioFilePath(fileName):
    speaker, batch = getSpeakerBatch(fileName)

    audioFilePath = "S{}/B{}/{}-%02d.wav".format(speaker, batch,
                                                 speaker + '-' + batch)

    audioFilePath = os.path.join(
        config.AUDIO_SPLIT_DIR, audioFilePath)

    return audioFilePath


def getSeparatedVideoFilePath(fileName):
    speaker, batch = getSpeakerBatch(fileName)

    videoFilePath = "S{}/B{}/{}-%02d.mp4".format(speaker, batch,
                                                 speaker + '-' + batch)

    videoFilePath = os.path.join(
        config.VIDEO_SPLIT_DIR, videoFilePath)

    return videoFilePath


def getMergedVideoFilePath(fileName):
    speaker, batch = getSpeakerBatch(fileName)

    videoFilePath = "S{}/B{}/{}-%02d.mp4".format(speaker, batch,
                                                 speaker + '-' + batch)

    videoFilePath = os.path.join(
        config.MERGE_SPLIT_DIR, videoFilePath)

    return videoFilePath



from AudioDetection.Audio import Audio
from interface import implements
import os
from config import AUDIO_DIR
from scipy.io import wavfile
from AudioDetection.audioSegmentation import silence_removal
from Utilities import Utilities as util


class WavAudio(implements(Audio)):
    def __init__(self, audioFileName):
        if len(audioFileName.split('/')) > 1:
            raise Exception(
                "Invalid audio file name. Valid example: 001-01.wav")

        self.audioFileName = audioFileName
        self.audioFilePath = util.getAudioFilePath(self.audioFileName)

        if not os.path.isfile(self.audioFilePath):
            raise FileNotFoundError(
                "Audio file '{}' not does not exists".format(self.audioFilePath))

    def read(self):
        self.samplingRate, self.audioData = wavfile.read(self.audioFilePath)
        return self.samplingRate, self.audioData

    def getAudioFileName(self):
        return self.audioFileName

    def findIntervals(self, smoothingWindow, weight, plot):
        intervals = silence_removal(
            self.audioData, self.samplingRate, 0.05, 0.05, smoothingWindow, weight, plot)
        return intervals



from AudioDetection.WavAudio import WavAudio
import scipy.io.wavfile as wavfile
from numpy import genfromtxt
from Utilities import Utilities as util
from tqdm import tqdm


class AudioSeparator:
    def __init__(self, audioFileName):
        self.audioFileName = audioFileName

    def apply(self):
        samplingRate, audioData = WavAudio(self.audioFileName).read()

        intervalFileName = util.getIntervalPath(self.audioFileName)
        audioFilePath = util.getSeparatedAudioFilePath(self.audioFileName)

        intervals = genfromtxt(intervalFileName, delimiter=',')

        for i in tqdm(range(len(intervals)), desc='Audio separation in progress'):
            wavfile.write(audioFilePath % (i + 1), samplingRate, audioData[int(
                samplingRate * intervals[i][0]): int(samplingRate * intervals[i][1])])



import cv2
from ffpyplayer.player import MediaPlayer
import config
from glob import glob
from Utilities import Utilities as util
from tqdm import tqdm


def play(fileName):
    video = cv2.VideoCapture(fileName)
    player = MediaPlayer(fileName)

    while True:
        grabbed, frame = video.read()
        audioFrame, val = player.get_frame()
        if not grabbed:
            break
        if cv2.waitKey(28) & 0xFF == ord("q"):
            break
        cv2.imshow("Video", frame)
        if val != 'eof' and audioFrame is not None:
            image, t = audioFrame
    video.release()
    cv2.destroyAllWindows()


def playAll(fileName):
    mergedVideoFilePath = util.getMergedVideoFilePath(fileName)

    for i in tqdm(range(config.WORDS_PER_BATCH)):
        play(mergedVideoFilePath % (i + 1))



import numpy as np
from scipy import linalg as la
from sklearn import svm

shortTermWindow = 0.050
shortTermStep = 0.050
eps = 0.00000001


def train_svm(features, c_param, kernel='linear'):
    """
    Train a multi-class probabilitistic SVM classifier.
    Note:     This function is simply a wrapper to the sklearn functionality 
              for SVM training
              See function trainSVM_feature() to use a wrapper on both the 
              feature extraction and the SVM training
              (and parameter tuning) processes.
    ARGUMENTS:
        - features:         a list ([numOfClasses x 1]) whose elements 
                            containt np matrices of features  each matrix 
                            features[i] of class i is 
                            [n_samples x numOfDimensions]
        - c_param:           SVM parameter C (cost of constraints violation)
    RETURNS:
        - svm:              the trained SVM variable

    NOTE:
        This function trains a linear-kernel SVM for a given C value.
        For a different kernel, other types of parameters should be provided.
    """

    feature_matrix, labels = features_to_matrix(features)
    _svm = svm.SVC(C=c_param, kernel=kernel, probability=True,
                   gamma='auto')
    _svm.fit(feature_matrix, labels)

    return _svm


def normalize_features(features):
    """
    This function normalizes a feature set to 0-mean and 1-std.
    Used in most classifier trainning cases.

    ARGUMENTS:
        - features:    list of feature matrices (each one of them is a np
                       matrix)
    RETURNS:
        - features_norm:    list of NORMALIZED feature matrices
        - mean:        mean vector
        - std:        std vector
    """
    temp_feats = np.array([])

    for count, f in enumerate(features):
        if f.shape[0] > 0:
            if count == 0:
                temp_feats = f
            else:
                temp_feats = np.vstack((temp_feats, f))
            count += 1

    mean = np.mean(temp_feats, axis=0) + 1e-14
    std = np.std(temp_feats, axis=0) + 1e-14

    features_norm = []
    for f in features:
        ft = f.copy()
        for n_samples in range(f.shape[0]):
            ft[n_samples, :] = (ft[n_samples, :] - mean) / std
        features_norm.append(ft)
    return features_norm, mean, std


def features_to_matrix(features):
    """
    features_to_matrix(features)

    This function takes a list of feature matrices as argument and returns
    a single concatenated feature matrix and the respective class labels.

    ARGUMENTS:
        - features:        a list of feature matrices

    RETURNS:
        - feature_matrix:    a concatenated matrix of features
        - labels:            a vector of class indices
    """

    labels = np.array([])
    feature_matrix = np.array([])
    for i, f in enumerate(features):
        if i == 0:
            feature_matrix = f
            labels = i * np.ones((len(f), 1))
        else:
            feature_matrix = np.vstack((feature_matrix, f))
            labels = np.append(labels, i * np.ones((len(f), 1)))
    return feature_matrix, labels



from config import INTERVAL_DIR, WORDS_PER_BATCH
from .Audio import Audio
from numpy import savetxt
import os
import scipy.io.wavfile as wavfile


class AudioDetector:
    def __init__(self, audio):
        self.audio: Audio = audio

    def saveAsCsv(self, intervals):
        csvFilePath = os.path.join(
            INTERVAL_DIR, self.audio.getAudioFileName().split('.')[0] + '.csv')

        if not os.path.isfile(csvFilePath):
            with open(csvFilePath, "w") as filePath:
                pass
        savetxt(csvFilePath, intervals, fmt='%.2f', delimiter=',')

    def apply(self, smoothingWindow=1.0, weight=0.3, plot=False):
        self.audio.read()
        intervals = self.audio.findIntervals(smoothingWindow, weight, plot)

        if len(intervals) != WORDS_PER_BATCH:
            print(
                'Error: In AudioDetector.py at line 26: Could not detect all the words.')

        self.saveAsCsv(intervals)



import subprocess
from Utilities import Utilities as util


class Compressor:
    def __init__(self, fileName):
        self.fileName = fileName

    def applySizeCompression(self, factor):
        filePath = util.getUncompressedVideoPath(self.fileName)
        targetPath = util.getVideoPath(self.fileName)
        compressCommand = 'ffmpeg -loglevel quiet -i {} -vf \"scale=iw/{}:ih/{}\" {}'.format(
            filePath, factor, factor, targetPath)
        subprocess.call(compressCommand, shell=True)



from os import write
from Utilities import Utilities as util
import cv2
import numpy as np
from tqdm import tqdm
from FaceDetection import FaceDetector as detector


class VideoSeparator:
    def __init__(self, videoFileName):
        self.videoFileName = videoFileName
        self.videoFilePath = util.getVideoPath(videoFileName)

        self.capture = cv2.VideoCapture(self.videoFilePath)
        self.fps = self.capture.get(cv2.CAP_PROP_FPS)
        self.size = (int(self.capture.get(cv2.CAP_PROP_FRAME_WIDTH)),
                     int(self.capture.get(cv2.CAP_PROP_FRAME_HEIGHT)))
        self.fourcc = int(cv2.VideoWriter_fourcc('m', 'p', '4', 'v'))

    def getFrameId(self, t):
        return round(t * self.fps)

    def apply(self):

        intervalFileName = util.getIntervalPath(self.videoFileName)
        videoFilePath = util.getSeparatedVideoFilePath(self.videoFileName)
        intervals = np.genfromtxt(intervalFileName, delimiter=',')

        for i in tqdm(range(len(intervals)), desc='Video separation progress'):
            start = intervals[i][0]
            end = intervals[i][1]

            beginFrameId = self.getFrameId(start)
            endFrameId = self.getFrameId(end)

            writer = cv2.VideoWriter(
                videoFilePath % (i + 1), self.fourcc, self.fps, self.size)

            self.capture.set(cv2.CAP_PROP_POS_FRAMES, beginFrameId)

            ret = True
            while(self.capture.isOpened() and ret and writer.isOpened()):
                ret, frame = self.capture.read()
                frameId = self.capture.get(cv2.CAP_PROP_POS_FRAMES) - 1
                if(frameId <= endFrameId):
                    writer.write(frame)
                else:
                    break
            writer.release()



from glob import glob
import config
import subprocess
from Utilities import Utilities as util
from tqdm import tqdm
import moviepy.editor as editor


class AudioVideoMerger:
    def __init__(self, fileName):
        self.fileName = fileName
        self.audioFilePath = util.getSeparatedAudioFilePath(self.fileName)
        self.videoFilePath = util.getSeparatedVideoFilePath(self.fileName)
        self.mergeVideoFilePath = util.getMergedVideoFilePath(self.fileName)

    def apply(self):
        for id in tqdm(range(config.WORDS_PER_BATCH), desc='Merge in progress:'):
            merge = 'ffmpeg -loglevel quiet -i {} -i {} -c:v copy -c:a aac {}'.format(
                self.videoFilePath % (id + 1), self.audioFilePath % (id + 1), self.mergeVideoFilePath % (id + 1))
            subprocess.call(merge, shell=True)

    def applyEditor(self):
        for id in tqdm(range(config.WORDS_PER_BATCH), desc='Merge in progress:'):
            video = editor.VideoFileClip(
                self.videoFilePath % (id + 1), audio_fps=False)

            audio = editor.AudioFileClip(self.audioFilePath % (
                id + 1), fps=config.DEFAULT_SAMPLING_RATE)

            audio = editor.CompositeAudioClip([audio])
            video = video.set_audio(audio)
            video.write_videofile(self.mergeVideoFilePath % (id + 1))



import numpy as np
import matplotlib.pyplot as plt
from scipy.spatial import distance
import AudioDetection.audioBasicIO as audioBasicIO
import AudioDetection.audioTrainTest as at
import AudioDetection.ShortTermFeatures as stf

""" General utility functions """


def smooth_moving_avg(signal, window=11):
    window = int(window)
    if signal.ndim != 1:
        raise ValueError("")
    if signal.size < window:
        raise ValueError("Input vector needs to be bigger than window size.")
    if window < 3:
        return signal
    s = np.r_[2 * signal[0] - signal[window - 1::-1],
              signal, 2 * signal[-1] - signal[-1:-window:-1]]
    w = np.ones(window, 'd')
    y = np.convolve(w/w.sum(), s, mode='same')
    return y[window:-window + 1]


def silence_removal(signal, sampling_rate, st_win, st_step, smooth_window=0.5,
                    weight=0.5, plot=False):  # default smooth_window = 0.5, weight=0.5
    """
    Event Detection (silence removal)
    ARGUMENTS:
         - signal:                the input audio signal
         - sampling_rate:               sampling freq
         - st_win, st_step:    window size and step in seconds
         - smoothWindow:     (optinal) smooth window (in seconds)
         - weight:           (optinal) weight factor (0 < weight < 1)
                              the higher, the more strict
         - plot:             (optinal) True if results are to be plotted
    RETURNS:
         - seg_limits:    list of segment limits in seconds (e.g [[0.1, 0.9],
                          [1.4, 3.0]] means that
                          the resulting segments are (0.1 - 0.9) seconds
                          and (1.4, 3.0) seconds
    """

    if weight >= 1:
        weight = 0.99
    if weight <= 0:
        weight = 0.01

    # Step 1: feature extraction
    signal = audioBasicIO.stereo_to_mono(signal)
    st_feats, _ = stf.feature_extraction(signal, sampling_rate,
                                         st_win * sampling_rate,
                                         st_step * sampling_rate)

    # Step 2: train binary svm classifier of low vs high energy frames
    # keep only the energy short-term sequence (2nd feature)
    st_energy = st_feats[1, :]
    en = np.sort(st_energy)
    # number of 10% of the total short-term windows
    st_windows_fraction = int(len(en) / 10)

    # compute "lower" 10% energy threshold
    low_threshold = np.mean(en[0:st_windows_fraction]) + 1e-15

    # compute "higher" 10% energy threshold
    high_threshold = np.mean(en[-st_windows_fraction:-1]) + 1e-15

    # get all features that correspond to low energy
    low_energy = st_feats[:, np.where(st_energy <= low_threshold)[0]]

    # get all features that correspond to high energy
    high_energy = st_feats[:, np.where(st_energy >= high_threshold)[0]]

    # form the binary classification task and ...
    features = [low_energy.T, high_energy.T]
    # normalize and train the respective svm probabilistic model

    # (ONSET vs SILENCE)
    features_norm, mean, std = at.normalize_features(features)
    svm = at.train_svm(features_norm, 1.0)

    # Step 3: compute onset probability based on the trained svm
    prob_on_set = []
    for index in range(st_feats.shape[1]):
        # for each frame
        cur_fv = (st_feats[:, index] - mean) / std
        # get svm probability (that it belongs to the ONSET class)
        prob_on_set.append(svm.predict_proba(cur_fv.reshape(1, -1))[0][1])
    prob_on_set = np.array(prob_on_set)

    # smooth probability:
    prob_on_set = smooth_moving_avg(prob_on_set, smooth_window / st_step)

    # Step 4A: detect onset frame indices:
    prog_on_set_sort = np.sort(prob_on_set)

    # find probability Threshold as a weighted average
    # of top 10% and lower 10% of the values
    nt = int(prog_on_set_sort.shape[0] / 10)
    threshold = (np.mean((1 - weight) * prog_on_set_sort[0:nt]) +
                 weight * np.mean(prog_on_set_sort[-nt::]))

    max_indices = np.where(prob_on_set > threshold)[0]
    # get the indices of the frames that satisfy the thresholding
    index = 0
    seg_limits = []
    time_clusters = []

    # Step 4B: group frame indices to onset segments
    while index < len(max_indices):
        # for each of the detected onset indices
        cur_cluster = [max_indices[index]]
        if index == len(max_indices)-1:
            break
        while max_indices[index+1] - cur_cluster[-1] <= 2:
            cur_cluster.append(max_indices[index+1])
            index += 1
            if index == len(max_indices)-1:
                break
        index += 1
        time_clusters.append(cur_cluster)
        seg_limits.append([cur_cluster[0] * st_step,
                           cur_cluster[-1] * st_step])

    # Step 5: Post process: remove very small segments:
    min_duration = 0.2
    seg_limits_2 = []
    for s_lim in seg_limits:
        if s_lim[1] - s_lim[0] > min_duration:
            seg_limits_2.append(s_lim)
    seg_limits = seg_limits_2

    if plot:
        time_x = np.arange(0, signal.shape[0] / float(sampling_rate), 1.0 /
                           sampling_rate)

        plt.subplot(2, 1, 1)
        plt.plot(time_x, signal)
        for s_lim in seg_limits:
            plt.axvline(x=s_lim[0], color='red')
            plt.axvline(x=s_lim[1], color='red')
        plt.subplot(2, 1, 2)
        plt.plot(np.arange(0, prob_on_set.shape[0] * st_step, st_step),
                 prob_on_set)
        plt.title('Signal')
        for s_lim in seg_limits:
            plt.axvline(x=s_lim[0], color='red')
            plt.axvline(x=s_lim[1], color='red')
        plt.title('svm Probability')
        plt.show()

    return seg_limits



from setuptools import setup
from pathlib import Path

root = Path(__file__).parent
with open(root/'README.md') as file:
    README = file.read()

REQUIREMENTS = [
    'autopep8==1.5.7',
    'certifi==2021.5.30',
    'charset-normalizer==2.0.3',
    'cycler==0.10.0',
    'decorator==4.4.2',
    'dlib==19.22.0',
    'ffmpeg==1.4',
    'ffmpeg-python==0.2.0',
    'ffpyplayer==4.3.2',
    'future==0.18.2',
    'idna==3.2',
    'imageio==2.9.0',
    'imageio-ffmpeg==0.4.4',
    'joblib==1.0.1',
    'kiwisolver==1.3.1',
    'matplotlib==3.4.2',
    'moviepy==1.0.3',
    'numpy==1.21.1',
    'opencv-python==4.5.3.56',
    'path==16.2.0',
    'Pillow==8.3.1',
    'proglog==0.1.9',
    'pycodestyle==2.7.0',
    'pyparsing==2.4.7',
    'python-dateutil==2.8.2',
    'python-interface==1.6.1',
    'requests==2.26.0',
    'scipy==1.7.0',
    'six==1.16.0',
    'threadpoolctl==2.2.0',
    'toml==0.10.2',
    'tqdm==4.61.2',
    'urllib3==1.26.6',
]

CLASSIFIERS = [
    'Development Status :: 4 - Beta',
    'Intended Audience :: Developers',
    'Topic :: Internet',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
]

setup(
    name='bhuiyans-dataset-builder',
    version='0.0.3',
    description='A video processing tool for visual data collection, end-to-end preprocessing, ready-to-go for model training.',
    long_description=README,
    url='https://github.com/MasumBhuiyan/dataset-builder',
    author='Masum Bhuiyan',
    author_email='masumbhuiyan577@gmail.com',
    license='Apache License',
    package_dir={'': 'src'},
    py_modules=['config', 'app'],
    include_package_data=True,
    packages=['AudioDetection',
              'AudioExtraction',
              'AudioPlayer',
              'AudioSeparation',
              'AudioVideoMerge',
              'Compression',
              'FaceDetection',
              'Preprocessing',
              'Utilities',
              'VideoPlayer',
              'VideoSeparation'],
    classifiers=CLASSIFIERS,
    install_requires=REQUIREMENTS,
    keywords='video preprocessing split merge video player'
)



import dlib
import cv2
import os
import config
from Utilities import Utilities as util
from glob import glob
from tqdm import tqdm

faceDetector = dlib.get_frontal_face_detector()
landmarkDetector = dlib.shape_predictor(
    config.LANDMARK_DETECTOR_PATH
)


def shape2List(landmark):
    points = []
    for i in range(48, 68):  # 0, 68 for all, 48, 68 for lip
        points.append((landmark.part(i).x, landmark.part(i).y))
    return points


def readGrayFrames(filePath):
    frames = []
    capture = cv2.VideoCapture(filePath)

    while True:
        success, frame = capture.read()
        if not success:
            break
        gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
        frames.append(gray)
    capture.release()
    return frames


def detectLandmark(file, frames):
    landmarks = []
    for i, frame in enumerate(frames):
        faces = faceDetector(frame, 1)

        if len(faces) < 1:
            print('Skipping {}: No face detected.'.format(file))
            if len(landmarks) > 0:
                landmarks.append(landmarks[-1])
            continue

        if len(faces) > 1:
            print('Skipping {}: Too many face detected'.format(file))
            if len(landmarks) > 0:
                landmarks.append(landmarks[-1])
            continue

        face = faces[0]
        landmark = landmarkDetector(frame, face)
        landmark = shape2List(landmark)
        landmarks.append(landmark)
    return landmarks


def extractLip(frames, landmarks):
    lips = []
    for i, landmark in enumerate(landmarks):
        # Landmark corresponding to lip
        lip = landmark

        # Lip landmark sorted for determining lip region
        lip_x = sorted(lip, key=lambda pointx: pointx[0])
        lip_y = sorted(lip, key=lambda pointy: pointy[1])

        # Determine Margins for lip-only image
        x_add = int((-lip_x[0][0]+lip_x[-1][0]) * config.LIP_MARGIN)
        y_add = int((-lip_y[0][1]+lip_y[-1][1]) * config.LIP_MARGIN)

        crop_pos = (lip_x[0][0]-x_add, lip_x[-1][0]+x_add,
                    lip_y[0][1]-y_add, lip_y[-1][1]+y_add)   # Crop image

        cropped = frames[i][crop_pos[2]:crop_pos[3], crop_pos[0]:crop_pos[1]]
        cropped = cv2.resize(
            cropped, (config.LIP_CROP_SIZE[0], config.LIP_CROP_SIZE[1]), interpolation=cv2.INTER_CUBIC)        # Resize

        lips.append(cropped)
    return lips


def save(targetPath, lips):
    for i, lip in enumerate(lips):
        cv2.imwrite(targetPath + "%03d" % (i + 1) + ".jpg", lip)


def apply(speaker):
    filePath = '{}/S{}//*-03-*.mp4'.format(config.MERGE_SPLIT_DIR, speaker)
    allFiles = sorted(glob(filePath))

    for i in tqdm(range(len(allFiles)), desc='Lip extraction in progress'):
        file = allFiles[i]
        frames = readGrayFrames(file)
        landmarks = detectLandmark(file, frames)
        lips = extractLip(frames, landmarks)

        fileName = file.split('/')[-1]
        speaker, batch = util.getSpeakerBatch(fileName)
        filePrefix = fileName.split('.')[0]

        targetPath = '{}/S{}/B{}/{}/'.format(
            config.PROCESSED_VIDEO_DIR, speaker, batch, filePrefix)

        if not os.path.exists(targetPath):
            os.makedirs(targetPath)

        save(targetPath, lips)



AUDIO_DIR = './asset/audio'
VIDEO_DIR = './asset/video'
INTERVAL_DIR = './asset/time'
UNCOMPRESSED_VIDEO_DIR = './asset/uncompressed-video'
PROCESSED_VIDEO_DIR = './asset/processed-video'
AUDIO_SPLIT_DIR = './asset/separated-audio'
VIDEO_SPLIT_DIR = './asset/separated-video'
WORDLIST_DIR = './asset/wordlists'
MERGE_SPLIT_DIR = './asset/merged-video'

LANDMARK_DETECTOR_PATH = './src/FaceDetection/shape_predictor_68_face_landmarks.dat'
LIP_MARGIN = 0.3
LIP_CROP_SIZE = (112, 112)

WORDS_PER_BATCH = 25
DEFAULT_SAMPLING_RATE = 44100  # Hz
SPEAKER_ID_LENGTH = 3
BATCH_ID_LENGTH = 2



import argparse
from pathlib import Path
from Compression.Compressor import Compressor
from AudioVideoMerge.AudioVideoMerger import AudioVideoMerger
from VideoSeparation.VideoSeparator import VideoSeparator
from config import AUDIO_SPLIT_DIR, VIDEO_SPLIT_DIR, MERGE_SPLIT_DIR
from AudioExtraction.AudioExtractor import AudioExtractor
from AudioExtraction.Mp4Video import Mp4Video
from AudioDetection.AudioDetector import AudioDetector
from AudioDetection.WavAudio import WavAudio
from AudioSeparation.AudioSeparator import AudioSeparator
import Utilities.Utilities as util
import VideoPlayer.VideoPlayer as player
from FaceDetection import FaceDetector as detector

parser = argparse.ArgumentParser()

parser.add_argument('--fileName')
parser.add_argument('--extractAudio')
parser.add_argument('--detectAudio')
parser.add_argument('--plot')
parser.add_argument('--separateAudio')
parser.add_argument('--separateVideo')
parser.add_argument('--mergeAudioVideo')
parser.add_argument('--playAll')
parser.add_argument('--compressBySize')
parser.add_argument('--execute')
parser.add_argument('--detectLip')
parser.add_argument('--speaker')


def extractAudio(videoFileName):
    audioExtractor = AudioExtractor(Mp4Video(videoFileName))
    audioExtractor.apply()


def detectAudio(audioFileName, plot):
    if plot == None:
        plot = False
    audioDetector = AudioDetector(WavAudio(audioFileName))
    audioDetector.apply(plot=plot)


def separateAudio(audioFileName):
    audioSeparator = AudioSeparator(audioFileName)
    audioSeparator.apply()


def separateVideo(videoFileName):
    videoSeparator = VideoSeparator(videoFileName)
    videoSeparator.apply()


def mergeAudioVideo(videoFileName):
    audioVideoMerger = AudioVideoMerger(videoFileName)
    audioVideoMerger.apply()


def playAll(fileName):
    print("playing..")
    player.playAll(fileName)


def compressBySize(fileName, factor):
    compressor = Compressor(fileName)
    compressor.applySizeCompression(factor)


def detectLip(speaker):
    detector.apply(speaker)


def addPadding():
    # add small time to front and back of asset/time/ files for smoothing
    pass


def createDirectories(fileName):
    speaker, batch = util.getSpeakerBatch(fileName)

    Path("{}/S{}/B{}".format(AUDIO_SPLIT_DIR, speaker, batch)
         ).mkdir(parents=True, exist_ok=True)

    Path("{}/S{}/B{}".format(VIDEO_SPLIT_DIR, speaker, batch)
         ).mkdir(parents=True, exist_ok=True)

    Path("{}/S{}/B{}".format(MERGE_SPLIT_DIR, speaker, batch)
         ).mkdir(parents=True, exist_ok=True)


if __name__ == "__main__":
    args = parser.parse_args()

    if(args.fileName == None):
        raise Exception(
            'Missing: Requires filename.')

    createDirectories(args.fileName)

    if args.extractAudio == 'yes':
        extractAudio(args.fileName)

    if args.detectAudio == 'yes':
        detectAudio(args.fileName, args.plot)

    if args.separateAudio == 'yes':
        separateAudio(args.fileName)

    if args.separateVideo == 'yes':
        separateVideo(args.fileName)

    if args.mergeAudioVideo == 'yes':
        mergeAudioVideo(args.fileName)

    if args.playAll == 'yes':
        playAll(args.fileName)

    if args.compressBySize != None:
        compressBySize(args.fileName, args.compressBySize)

    if args.execute == 'all':
        extractAudio(args.fileName)
        detectAudio(args.fileName, args.plot)
        separateAudio(args.fileName)
        separateVideo(args.fileName)
        mergeAudioVideo(args.fileName)
        playAll(args.fileName)

    if args.detectLip == 'yes':
        if args.speaker == None:
            raise Exception('Requires speaker id for detecting lip regions.')
        detectLip(args.speaker)



import numpy as np
import sys
from scipy.fftpack import fft
from scipy.fftpack.realtransforms import dct

eps = sys.float_info.epsilon


def dc_normalize(sig_array):
    """Removes DC and normalizes to -1, 1 range"""
    sig_array_norm = sig_array.copy()
    sig_array_norm -= sig_array_norm.mean()
    sig_array_norm /= abs(sig_array_norm).max() + 1e-10
    return sig_array_norm


def zero_crossing_rate(frame):
    """Computes zero crossing rate of frame"""
    count = len(frame)
    count_zero = np.sum(np.abs(np.diff(np.sign(frame)))) / 2
    return np.float64(count_zero) / np.float64(count - 1.0)


def energy(frame):
    """Computes signal energy of frame"""
    return np.sum(frame ** 2) / np.float64(len(frame))


def energy_entropy(frame, n_short_blocks=10):
    """Computes entropy of energy"""
    # total frame energy
    frame_energy = np.sum(frame ** 2)
    frame_length = len(frame)
    sub_win_len = int(np.floor(frame_length / n_short_blocks))
    if frame_length != sub_win_len * n_short_blocks:
        frame = frame[0:sub_win_len * n_short_blocks]

    # sub_wins is of size [n_short_blocks x L]
    sub_wins = frame.reshape(sub_win_len, n_short_blocks, order='F').copy()

    # Compute normalized sub-frame energies:
    s = np.sum(sub_wins ** 2, axis=0) / (frame_energy + eps)

    # Compute entropy of the normalized sub-frame energies:
    entropy = -np.sum(s * np.log2(s + eps))
    return entropy


""" Frequency-domain audio features """


def spectral_centroid_spread(fft_magnitude, sampling_rate):
    """Computes spectral centroid of frame (given abs(FFT))"""
    ind = (np.arange(1, len(fft_magnitude) + 1)) * \
          (sampling_rate / (2.0 * len(fft_magnitude)))

    Xt = fft_magnitude.copy()
    Xt_max = Xt.max()
    if Xt_max == 0:
        Xt = Xt / eps
    else:
        Xt = Xt / Xt_max

    NUM = np.sum(ind * Xt)
    DEN = np.sum(Xt) + eps

    # Centroid:
    centroid = (NUM / DEN)

    # Spread:
    spread = np.sqrt(np.sum(((ind - centroid) ** 2) * Xt) / DEN)

    # Normalize:
    centroid = centroid / (sampling_rate / 2.0)
    spread = spread / (sampling_rate / 2.0)

    return centroid, spread


def spectral_entropy(signal, n_short_blocks=10):
    """Computes the spectral entropy"""
    # number of frame samples
    num_frames = len(signal)

    # total spectral energy
    total_energy = np.sum(signal ** 2)

    # length of sub-frame
    sub_win_len = int(np.floor(num_frames / n_short_blocks))
    if num_frames != sub_win_len * n_short_blocks:
        signal = signal[0:sub_win_len * n_short_blocks]

    # define sub-frames (using matrix reshape)
    sub_wins = signal.reshape(sub_win_len, n_short_blocks, order='F').copy()

    # compute spectral sub-energies
    s = np.sum(sub_wins ** 2, axis=0) / (total_energy + eps)

    # compute spectral entropy
    entropy = -np.sum(s * np.log2(s + eps))

    return entropy


def spectral_flux(fft_magnitude, previous_fft_magnitude):
    """
    Computes the spectral flux feature of the current frame
    ARGUMENTS:
        fft_magnitude:            the abs(fft) of the current frame
        previous_fft_magnitude:        the abs(fft) of the previous frame
    """
    # compute the spectral flux as the sum of square distances:
    fft_sum = np.sum(fft_magnitude + eps)
    previous_fft_sum = np.sum(previous_fft_magnitude + eps)
    sp_flux = np.sum(
        (fft_magnitude / fft_sum - previous_fft_magnitude /
         previous_fft_sum) ** 2)

    return sp_flux


def spectral_rolloff(signal, c):
    """Computes spectral roll-off"""
    energy = np.sum(signal ** 2)
    fft_length = len(signal)
    threshold = c * energy
    # Ffind the spectral rolloff as the frequency position
    # where the respective spectral energy is equal to c*totalEnergy
    cumulative_sum = np.cumsum(signal ** 2) + eps
    a = np.nonzero(cumulative_sum > threshold)[0]
    if len(a) > 0:
        sp_rolloff = np.float64(a[0]) / (float(fft_length))
    else:
        sp_rolloff = 0.0
    return sp_rolloff


def mfcc(fft_magnitude, fbank, num_mfcc_feats):
    """
    Computes the MFCCs of a frame, given the fft mag

    ARGUMENTS:
        fft_magnitude:  fft magnitude abs(FFT)
        fbank:          filter bank (see mfccInitFilterBanks)
    RETURN
        ceps:           MFCCs (13 element vector)

    Note:    MFCC calculation is, in general, taken from the 
             scikits.talkbox library (MIT Licence),
    #    with a small number of modifications to make it more 
         compact and suitable for the pyAudioAnalysis Lib
    """

    mspec = np.log10(np.dot(fft_magnitude, fbank.T) + eps)
    ceps = dct(mspec, type=2, norm='ortho', axis=-1)[:num_mfcc_feats]
    return ceps


def chroma_features_init(num_fft, sampling_rate):
    """
    This function initializes the chroma matrices used in the calculation
    of the chroma features
    """
    freqs = np.array([((f + 1) * sampling_rate) /
                      (2 * num_fft) for f in range(num_fft)])
    cp = 27.50
    num_chroma = np.round(12.0 * np.log2(freqs / cp)).astype(int)

    num_freqs_per_chroma = np.zeros((num_chroma.shape[0],))

    unique_chroma = np.unique(num_chroma)
    for u in unique_chroma:
        idx = np.nonzero(num_chroma == u)
        num_freqs_per_chroma[idx] = idx[0].shape

    return num_chroma, num_freqs_per_chroma


def chroma_features(signal, sampling_rate, num_fft):
    # TODO: 1 complexity
    # TODO: 2 bug with large windows

    num_chroma, num_freqs_per_chroma = \
        chroma_features_init(num_fft, sampling_rate)
    chroma_names = ['A', 'A#', 'B', 'C', 'C#', 'D',
                    'D#', 'E', 'F', 'F#', 'G', 'G#']
    spec = signal ** 2
    if num_chroma.max() < num_chroma.shape[0]:
        C = np.zeros((num_chroma.shape[0],))
        C[num_chroma] = spec
        C /= num_freqs_per_chroma[num_chroma]
    else:
        I = np.nonzero(num_chroma > num_chroma.shape[0])[0][0]
        C = np.zeros((num_chroma.shape[0],))
        C[num_chroma[0:I - 1]] = spec
        C /= num_freqs_per_chroma
    final_matrix = np.zeros((12, 1))
    newD = int(np.ceil(C.shape[0] / 12.0) * 12)
    C2 = np.zeros((newD,))
    C2[0:C.shape[0]] = C
    C2 = C2.reshape(int(C2.shape[0] / 12), 12)
    # for i in range(12):
    #    finalC[i] = np.sum(C[i:C.shape[0]:12])
    final_matrix = np.matrix(np.sum(C2, axis=0)).T

    spec_sum = spec.sum()
    if spec_sum == 0:
        final_matrix /= eps
    else:
        final_matrix /= spec_sum

    #    ax = plt.gca()
    #    plt.hold(False)
    #    plt.plot(finalC)
    #    ax.set_xticks(range(len(chromaNames)))
    #    ax.set_xticklabels(chromaNames)
    #    xaxis = np.arange(0, 0.02, 0.01);
    #    ax.set_yticks(range(len(xaxis)))
    #    ax.set_yticklabels(xaxis)
    #    plt.show(block=False)
    #    plt.draw()

    return chroma_names, final_matrix


def mfcc_filter_banks(sampling_rate, num_fft, lowfreq=133.33, linc=200 / 3,
                      logsc=1.0711703, num_lin_filt=13, num_log_filt=27):
    """
    Computes the triangular filterbank for MFCC computation 
    (used in the stFeatureExtraction function before the stMFCC function call)
    This function is taken from the scikits.talkbox library (MIT Licence):
    https://pypi.python.org/pypi/scikits.talkbox
    """

    if sampling_rate < 8000:
        nlogfil = 5

    # Total number of filters
    num_filt_total = num_lin_filt + num_log_filt

    # Compute frequency points of the triangle:
    frequencies = np.zeros(num_filt_total + 2)
    frequencies[:num_lin_filt] = lowfreq + np.arange(num_lin_filt) * linc
    frequencies[num_lin_filt:] = frequencies[num_lin_filt - 1] * logsc ** \
        np.arange(1, num_log_filt + 3)
    heights = 2. / (frequencies[2:] - frequencies[0:-2])

    # Compute filterbank coeff (in fft domain, in bins)
    fbank = np.zeros((num_filt_total, num_fft))
    nfreqs = np.arange(num_fft) / (1. * num_fft) * sampling_rate

    for i in range(num_filt_total):
        low_freqs = frequencies[i]
        cent_freqs = frequencies[i + 1]
        high_freqs = frequencies[i + 2]

        lid = np.arange(np.floor(low_freqs * num_fft / sampling_rate) + 1,
                        np.floor(cent_freqs * num_fft / sampling_rate) + 1,
                        dtype=np.int)
        lslope = heights[i] / (cent_freqs - low_freqs)
        rid = np.arange(np.floor(cent_freqs * num_fft / sampling_rate) + 1,
                        np.floor(high_freqs * num_fft / sampling_rate) + 1,
                        dtype=np.int)
        rslope = heights[i] / (high_freqs - cent_freqs)
        fbank[i][lid] = lslope * (nfreqs[lid] - low_freqs)
        fbank[i][rid] = rslope * (high_freqs - nfreqs[rid])

    return fbank, frequencies


""" Windowing and feature extraction """


def feature_extraction(signal, sampling_rate, window, step, deltas=True):
    """
    This function implements the shor-term windowing process.
    For each short-term window a set of features is extracted.
    This results to a sequence of feature vectors, stored in a np matrix.

    ARGUMENTS
        signal:         the input signal samples
        sampling_rate:  the sampling freq (in Hz)
        window:         the short-term window size (in samples)
        step:           the short-term window step (in samples)
        deltas:         (opt) True/False if delta features are to be
                        computed
    RETURNS
        features (numpy.ndarray):        contains features
                                         (n_feats x numOfShortTermWindows)
        feature_names (python list):     contains feature names
                                         (n_feats x numOfShortTermWindows)
    """

    window = int(window)
    step = int(step)

    # signal normalization
    signal = np.double(signal)
    signal = signal / (2.0 ** 15)

    signal = dc_normalize(signal)

    number_of_samples = len(signal)  # total number of samples
    current_position = 0
    count_fr = 0
    num_fft = int(window / 2)

    # compute the triangular filter banks used in the mfcc calculation
    fbank, freqs = mfcc_filter_banks(sampling_rate, num_fft)

    n_time_spectral_feats = 8
    n_harmonic_feats = 0
    n_mfcc_feats = 13
    n_chroma_feats = 13
    n_total_feats = n_time_spectral_feats + n_mfcc_feats + n_harmonic_feats + \
        n_chroma_feats
    #    n_total_feats = n_time_spectral_feats + n_mfcc_feats +
    #    n_harmonic_feats

    # define list of feature names
    feature_names = ["zcr", "energy", "energy_entropy"]
    feature_names += ["spectral_centroid", "spectral_spread"]
    feature_names.append("spectral_entropy")
    feature_names.append("spectral_flux")
    feature_names.append("spectral_rolloff")
    feature_names += ["mfcc_{0:d}".format(mfcc_i)
                      for mfcc_i in range(1, n_mfcc_feats + 1)]
    feature_names += ["chroma_{0:d}".format(chroma_i)
                      for chroma_i in range(1, n_chroma_feats)]
    feature_names.append("chroma_std")

    # add names for delta features:
    if deltas:
        feature_names_2 = feature_names + ["delta " + f for f in feature_names]
        feature_names = feature_names_2

    features = []
    # for each short-term window to end of signal
    while current_position + window - 1 < number_of_samples:
        count_fr += 1
        # get current window
        x = signal[current_position:current_position + window]

        # update window position
        current_position = current_position + step

        # get fft magnitude
        fft_magnitude = abs(fft(x))

        # normalize fft
        fft_magnitude = fft_magnitude[0:num_fft]
        fft_magnitude = fft_magnitude / len(fft_magnitude)

        # keep previous fft mag (used in spectral flux)
        if count_fr == 1:
            fft_magnitude_previous = fft_magnitude.copy()
        feature_vector = np.zeros((n_total_feats, 1))

        # zero crossing rate
        feature_vector[0] = zero_crossing_rate(x)

        # short-term energy
        feature_vector[1] = energy(x)

        # short-term entropy of energy
        feature_vector[2] = energy_entropy(x)

        # sp centroid/spread
        [feature_vector[3], feature_vector[4]] = \
            spectral_centroid_spread(fft_magnitude,
                                     sampling_rate)

        # spectral entropy
        feature_vector[5] = \
            spectral_entropy(fft_magnitude)

        # spectral flux
        feature_vector[6] = \
            spectral_flux(fft_magnitude,
                          fft_magnitude_previous)

        # spectral rolloff
        feature_vector[7] = \
            spectral_rolloff(fft_magnitude, 0.90)

        # MFCCs
        mffc_feats_end = n_time_spectral_feats + n_mfcc_feats
        feature_vector[n_time_spectral_feats:mffc_feats_end, 0] = \
            mfcc(fft_magnitude, fbank, n_mfcc_feats).copy()

        # chroma features
        chroma_names, chroma_feature_matrix = \
            chroma_features(fft_magnitude, sampling_rate, num_fft)
        chroma_features_end = n_time_spectral_feats + n_mfcc_feats + \
            n_chroma_feats - 1
        feature_vector[mffc_feats_end:chroma_features_end] = \
            chroma_feature_matrix
        feature_vector[chroma_features_end] = chroma_feature_matrix.std()
        if not deltas:
            features.append(feature_vector)
        else:
            # delta features
            if count_fr > 1:
                delta = feature_vector - feature_vector_prev
                feature_vector_2 = np.concatenate((feature_vector, delta))
            else:
                feature_vector_2 = np.concatenate((feature_vector,
                                                   np.zeros(feature_vector.
                                                            shape)))
            feature_vector_prev = feature_vector
            features.append(feature_vector_2)

        fft_magnitude_previous = fft_magnitude.copy()

    features = np.concatenate(features, 1)
    return features, feature_names
